@@ -776,16 +776,9 @@ normalizeConFor reifiedDec typename params variant = fmap (map giveTyVarBndrsSta
                        -> ConstructorVariant
                        -> Q [ConstructorInfo]
           dataFamCase' n tyvars stricts variant = do
-            info <- reifyRecover n $ fail $ unlines
-                      [ "normalizeCon: Cannot reify constructor " ++ nameBase n
-                      , "You are likely calling normalizeDec on GHC 7.6 or 7.8 on a data family"
-                      , "whose type variables have been eta-reduced due to GHC Trac #9692."
-                      , "Unfortunately, without being able to reify the constructor's type,"
-                      , "there is no way to recover the eta-reduced type variables in general."
-                      , "A recommended workaround is to use reifyDatatype instead."
-                      ]
-            case info of
-              DataConI _ ty _ _ -> do
+            mbInfo <- reifyMaybe n
+            case mbInfo of
+              Just (DataConI _ ty _ _) -> do
                 let (context, argTys :|- returnTy) = uncurryType ty
                 returnTy' <- resolveTypeSynonyms returnTy
                 -- Notice that we've ignored the Cxt and argument Types from the
@@ -800,7 +793,14 @@ normalizeConFor reifiedDec typename params variant = fmap (map giveTyVarBndrsSta
                 -- much easier.
                 normalizeGadtC typename params tyvars context [n]
                                returnTy' argTys stricts (const $ return variant)
-              _ -> fail "normalizeCon: impossible"
+              _ -> fail $ unlines
+                     [ "normalizeCon: Cannot reify constructor " ++ nameBase n
+                     , "You are likely calling normalizeDec on GHC 7.6 or 7.8 on a data family"
+                     , "whose type variables have been eta-reduced due to GHC Trac #9692."
+                     , "Unfortunately, without being able to reify the constructor's type,"
+                     , "there is no way to recover the eta-reduced type variables in general."
+                     , "A recommended workaround is to use reifyDatatype instead."
+                     ]
 
           -- A very ad hoc way of determining if we need to perform some extra passes
           -- to repair an eta-reduction bug for data family instances that only occurs
@@ -909,13 +909,17 @@ normalizeGadtC typename params tyvars context names innerType
      case decomposeType innerType' of
        ConT innerTyCon :| ts | typename == innerTyCon ->
 
-         let (substName, context1) = mergeArguments params ts
-             subst   = VarT <$> substName
-             tyvars' = [ tv | tv <- renamedTyvars, Map.notMember (tvName tv) subst ]
+         let (substName, context1) =
+               closeOverKinds (kindsOfFVsOfTvbs renamedTyvars)
+                              (kindsOfFVsOfTypes params)
+                              (mergeArguments params ts)
+             subst    = VarT <$> substName
+             exTyvars = [ tv | tv <- renamedTyvars, Map.notMember (tvName tv) subst ]
 
-             context2 = applySubstitution subst (context1 ++ renamedContext)
-             fields'  = applySubstitution subst renamedFields
-         in sequence [ ConstructorInfo name tyvars' context2
+             exTyvars' = substTyVarBndrs   subst exTyvars
+             context2  = applySubstitution subst (context1 ++ renamedContext)
+             fields'   = applySubstitution subst renamedFields
+         in sequence [ ConstructorInfo name exTyvars' context2
                                        fields' stricts <$> variantQ
                      | name <- names
                      , let variantQ = getVariant name
@@ -923,25 +927,163 @@ normalizeGadtC typename params tyvars context names innerType
 
        _ -> fail "normalizeGadtC: Expected type constructor application"
 
+{-
+Extend a type variable renaming subtitution and a list of equality
+predicates by looking into kind information as much as possible.
+
+Why is this necessary? Consider the following example:
+
+  data (a1 :: k1) :~: (b1 :: k1) where
+    Refl :: forall k2 (a2 :: k2). a2 :~: a2
+
+After an initial call to mergeArguments, we will have the following
+substitution and context:
+
+* Substitution: [a2 :-> a1]
+* Context: (a2 ~ b1)
+
+We shouldn't stop there, however! We determine the existentially quantified
+type variables of a constructor by filtering out those constructor-bound
+variables which do not appear in the substitution that mergeArguments
+returns. In this example, Refl's bound variables are k2 and a2. a2 appears
+in the returned substitution, but k2 does not, which means that we would
+mistakenly conclude that k2 is existential!
+
+Although we don't have the full power of kind inference to guide us here, we
+can at least do the next best thing. Generally, the datatype-bound type
+variables and the constructor type variable binders contain all of the kind
+information we need, so we proceed as follows:
+
+1. Construct a map from each constructor-bound variable to its kind. (Do the
+   same for each datatype-bound variable). These maps are the first and second
+   arguments to closeOverKinds, respectively.
+2. Call mergeArguments once on the GADT return type and datatype-bound types,
+   and pass that in as the third argument to closeOverKinds.
+3. For each name-name pair in the supplied substitution, check if the first and
+   second names map to kinds in the first and second kind maps in
+   closeOverKinds, respectively. If so, associate the first kind with the
+   second kind.
+4. For each kind association discovered in part (3), call mergeArguments
+   on the lists of kinds. This will yield a kind substitution and kind
+   equality context.
+5. If the kind substitution is non-empty, then go back to step (3) and repeat
+   the process on the new kind substitution and context.
+
+   Otherwise, if the kind substitution is empty, then we have reached a fixed-
+   point (i.e., we have closed over the kinds), so proceed.
+6. Union up all of the substitutions and contexts, and return those.
+
+This algorithm is not perfect, as it will only catch everything if all of
+the kinds are explicitly mentioned somewhere (and not left quantified
+implicitly). Thankfully, reifying data types via Template Haskell tends to
+yield a healthy amount of kind signatures, so this works quite well in
+practice.
+-}
+closeOverKinds :: Map Name Kind
+               -> Map Name Kind
+               -> (Map Name Name, Cxt)
+               -> (Map Name Name, Cxt)
+closeOverKinds domainFVKinds rangeFVKinds = go
+  where
+    go :: (Map Name Name, Cxt) -> (Map Name Name, Cxt)
+    go (subst, context) =
+      let substList = Map.toList subst
+          (kindsInner, kindsOuter) =
+            unzip $
+            mapMaybe (\(d, r) -> do d' <- Map.lookup d domainFVKinds
+                                    r' <- Map.lookup r rangeFVKinds
+                                    return (d', r'))
+                     substList
+          (kindSubst, kindContext) = mergeArgumentKinds kindsOuter kindsInner
+          (restSubst, restContext)
+            = if Map.null kindSubst -- Fixed-point calculation
+                 then (Map.empty, [])
+                 else go (kindSubst, kindContext)
+          finalSubst   = Map.unions [subst, kindSubst, restSubst]
+          finalContext = nub $ concat [context, kindContext, restContext]
+            -- Use `nub` here in an effort to minimize the number of
+            -- redundant equality constraints in the returned context.
+      in (finalSubst, finalContext)
+
+-- Look into a list of types and map each free variable name to its kind.
+kindsOfFVsOfTypes :: [Type] -> Map Name Kind
+kindsOfFVsOfTypes = foldMap go
+  where
+    go :: Type -> Map Name Kind
+    go (ForallT {}) = error "`forall` type used in data family pattern"
+    go (AppT t1 t2) = go t1 `Map.union` go t2
+    go (SigT t k) =
+      let kSigs =
+#if MIN_VERSION_template_haskell(2,8,0)
+                  go k
+#else
+                  Map.empty
+#endif
+      in case t of
+           VarT n -> Map.insert n k kSigs
+           _      -> go t `Map.union` kSigs
+    go _ = Map.empty
+
+-- Look into a list of type variable binder and map each free variable name
+-- to its kind (also map the names that KindedTVs bind to their respective
+-- kinds). This function considers the kind of a PlainTV to be *.
+kindsOfFVsOfTvbs :: [TyVarBndr] -> Map Name Kind
+kindsOfFVsOfTvbs = foldMap go
+  where
+    go :: TyVarBndr -> Map Name Kind
+    go (PlainTV n) = Map.singleton n starK
+    go (KindedTV n k) =
+      let kSigs =
+#if MIN_VERSION_template_haskell(2,8,0)
+                  kindsOfFVsOfTypes [k]
+#else
+                  Map.empty
+#endif
+      in Map.insert n k kSigs
+
 mergeArguments ::
   [Type] {- ^ outer parameters                    -} ->
   [Type] {- ^ inner parameters (specializations ) -} ->
   (Map Name Name, Cxt)
 mergeArguments ns ts = foldr aux (Map.empty, []) (zip ns ts)
   where
-    aux (SigT x _, y) sc = aux (x,y) sc -- learn about kinds??
-    aux (x, SigT y _) sc = aux (x,y) sc
 
     aux (f `AppT` x, g `AppT` y) sc =
       aux (x,y) (aux (f,g) sc)
 
     aux (VarT n,p) (subst, context) =
       case p of
-        VarT m | Map.notMember m subst -> (Map.insert m n subst, context)
+        VarT m | m == n  -> (subst, context)
+                   -- If the two variables are the same, don't bother extending
+                   -- the substitution. (This is purely an optimization.)
+               | Just n' <- Map.lookup m subst
+               , n == n' -> (subst, context)
+                   -- If a variable is already in a substitution and it maps
+                   -- to the variable that we are trying to unify with, then
+                   -- leave the context alone. (Not doing so caused #46.)
+               | Map.notMember m subst -> (Map.insert m n subst, context)
         _ -> (subst, equalPred (VarT n) p : context)
 
+    aux (SigT x _, y) sc = aux (x,y) sc -- learn about kinds??
+    -- This matches *after* VarT so that we can compute a substitution
+    -- that includes the kind signature.
+    aux (x, SigT y _) sc = aux (x,y) sc
+
     aux _ sc = sc
 
+-- | A specialization of 'mergeArguments' to 'Kind'.
+-- Needed only for backwards compatibility with older versions of
+-- @template-haskell@.
+mergeArgumentKinds ::
+  [Kind] ->
+  [Kind] ->
+  (Map Name Name, Cxt)
+#if MIN_VERSION_template_haskell(2,8,0)
+mergeArgumentKinds = mergeArguments
+#else
+mergeArgumentKinds _ _ = (Map.empty, [])
+#endif
+
 -- | Expand all of the type synonyms in a type.
 resolveTypeSynonyms :: Type -> Q Type
 resolveTypeSynonyms t =
@@ -951,10 +1093,9 @@ resolveTypeSynonyms t =
 
   case f of
     ConT n ->
-      do info <- reifyRecover n $ fail
-                   "resolveTypeSynonyms: Cannot reify type synonym information"
-         case info of
-           TyConI (TySynD _ synvars def)
+      do mbInfo <- reifyMaybe n
+         case mbInfo of
+           Just (TyConI (TySynD _ synvars def))
              -> resolveTypeSynonyms $ expandSynonymRHS synvars xs def
            _ -> notTypeSynCase
     _ -> notTypeSynCase
@@ -976,10 +1117,9 @@ resolvePredSynonyms :: Pred -> Q Pred
 resolvePredSynonyms = resolveTypeSynonyms
 #else
 resolvePredSynonyms (ClassP n ts) = do
-  info <- reifyRecover n $ fail
-            "resolvePredSynonyms: Cannot reify type synonym information"
-  case info of
-    TyConI (TySynD _ synvars def)
+  mbInfo <- reifyMaybe n
+  case mbInfo of
+    Just (TyConI (TySynD _ synvars def))
       -> resolvePredSynonyms $ typeToPred $ expandSynonymRHS synvars ts def
     _ -> ClassP n <$> mapM resolveTypeSynonyms ts
 resolvePredSynonyms (EqualP t1 t2) = do
@@ -1170,7 +1310,30 @@ freshenFreeVariables t =
 
 -- | Class for types that support type variable substitution.
 class TypeSubstitution a where
-  -- | Apply a type variable substitution
+  -- | Apply a type variable substitution.
+  --
+  -- Note that 'applySubstitution' is /not/ capture-avoiding. To illustrate
+  -- this, observe that if you call this function with the following
+  -- substitution:
+  --
+  -- * @b :-> a@
+  --
+  -- On the following 'Type':
+  --
+  -- * @forall a. b@
+  --
+  -- Then it will return:
+  --
+  -- * @forall a. a@
+  --
+  -- However, because the same @a@ type variable was used in the range of the
+  -- substitution as was bound by the @forall@, the substituted @a@ is now
+  -- captured by the @forall@, resulting in a completely different function.
+  --
+  -- For @th-abstraction@'s purposes, this is acceptable, as it usually only
+  -- deals with globally unique type variable 'Name's. If you use
+  -- 'applySubstitution' in a context where the 'Name's aren't globally unique,
+  -- however, be aware of this potential problem.
   applySubstitution :: Map Name Type -> a -> a
   -- | Compute the free type variables
   freeVariables     :: a -> [Name]
@@ -1248,6 +1411,14 @@ instance TypeSubstitution Kind where
   applySubstitution _ k = k
 #endif
 
+-- | Substitutes into the kinds of type variable binders.
+-- Not capture-avoiding.
+substTyVarBndrs :: Map Name Type -> [TyVarBndr] -> [TyVarBndr]
+substTyVarBndrs subst = map go
+  where
+    go tvb@(PlainTV {}) = tvb
+    go (KindedTV n k)   = KindedTV n (applySubstitution subst k)
+
 ------------------------------------------------------------------------
 
 combineSubstitutions :: Map Name Type -> Map Name Type -> Map Name Type
@@ -1255,6 +1426,10 @@ combineSubstitutions x y = Map.union (fmap (applySubstitution y) x) y
 
 -- | Compute the type variable substitution that unifies a list of types,
 -- or fail in 'Q'.
+--
+-- All infix issue should be resolved before using 'unifyTypes'
+--
+-- Alpha equivalent quantified types are not unified.
 unifyTypes :: [Type] -> Q (Map Name Type)
 unifyTypes [] = return Map.empty
 unifyTypes (t:ts) =
@@ -1277,20 +1452,23 @@ unify' :: Type -> Type -> Either (Type,Type) (Map Name Type)
 
 unify' (VarT n) (VarT m) | n == m = pure Map.empty
 unify' (VarT n) t | n `elem` freeVariables t = Left (VarT n, t)
-                  | otherwise                = pure (Map.singleton n t)
+                  | otherwise                = Right (Map.singleton n t)
 unify' t (VarT n) | n `elem` freeVariables t = Left (VarT n, t)
-                  | otherwise                = pure (Map.singleton n t)
-
-unify' (ConT n) (ConT m) | n == m = pure Map.empty
+                  | otherwise                = Right (Map.singleton n t)
 
 unify' (AppT f1 x1) (AppT f2 x2) =
   do sub1 <- unify' f1 f2
      sub2 <- unify' (applySubstitution sub1 x1) (applySubstitution sub1 x2)
-     return (combineSubstitutions sub1 sub2)
+     Right (combineSubstitutions sub1 sub2)
 
-unify' (TupleT n) (TupleT m) | n == m = pure Map.empty
+-- Doesn't unify kind signatures
+unify' (SigT t _) u = unify' t u
+unify' t (SigT u _) = unify' t u
 
-unify' t u = Left (t,u)
+-- only non-recursive cases should remain at this point
+unify' t u
+  | t == u    = Right Map.empty
+  | otherwise = Left (t,u)
 
 
 -- | Construct an equality constraint. The implementation of 'Pred' varies
@@ -1313,7 +1491,6 @@ classPred =
   ClassP
 #endif
 
-
 -- | Match a 'Pred' representing an equality constraint. Returns
 -- arguments to the equality constraint if successful.
 asEqualPred :: Pred -> Maybe (Type,Type)
@@ -1506,9 +1683,7 @@ reifyFixityCompat n = recover (return Nothing) $
        _                     -> Nothing
 #endif
 
--- | Call 'reify' with an action to take if reification fails.
-reifyRecover ::
-  Name ->
-  Q Info {- ^ handle failure -} ->
-  Q Info
-reifyRecover n failure = failure `recover` reify n
+-- | Call 'reify' and return @'Just' info@ if successful or 'Nothing' if
+-- reification failed.
+reifyMaybe :: Name -> Q (Maybe Info)
+reifyMaybe n = return Nothing `recover` fmap Just (reify n)
