@@ -26,10 +26,8 @@ module Main (main) where
 import           Control.Monad (zipWithM_)
 #endif
 
-#if MIN_VERSION_template_haskell(2,8,0)
 import           Control.Monad (unless)
 import qualified Data.Map as Map
-#endif
 
 #if MIN_VERSION_base(4,7,0)
 import           Data.Type.Equality ((:~:)(..))
@@ -55,6 +53,9 @@ main =
      voidstosTest
      strictDemoTest
      recordVanillaTest
+#if MIN_VERSION_template_haskell(2,6,0)
+     t43Test
+#endif
 #if MIN_VERSION_template_haskell(2,7,0)
      dataFamilyTest
      ghc78bugTest
@@ -64,6 +65,7 @@ main =
      famLocalDecTest1
      famLocalDecTest2
      recordFamTest
+     t46Test
 #endif
      fixityLookupTest
 #if __GLASGOW_HASKELL__ >= 704
@@ -77,6 +79,11 @@ main =
 #if MIN_VERSION_template_haskell(2,8,0)
      kindSubstTest
 #endif
+#if __GLASGOW_HASKELL__ >= 800
+     t37Test
+     polyKindedExTyvarTest
+#endif
+     regressionTest44
 
 adt1Test :: IO ()
 adt1Test =
@@ -215,7 +222,7 @@ showableTest =
            , datatypeCons    =
                [ ConstructorInfo
                    { constructorName       = 'Showable
-                   , constructorVars       = [PlainTV a]
+                   , constructorVars       = [KindedTV a starK]
                    , constructorContext    = [classPred ''Show [VarT a]]
                    , constructorFields     = [VarT a]
                    , constructorStrictness = [notStrictAnnot]
@@ -319,6 +326,47 @@ recordVanillaTest =
   $(do info <- reifyRecord 'gadtrec1a
        validateCI info gadtRecVanillaCI)
 
+#if MIN_VERSION_template_haskell(2,6,0)
+t43Test :: IO ()
+t43Test =
+  $(do [decPlain] <- [d| data T43Plain where MkT43Plain :: T43Plain |]
+       infoPlain  <- normalizeDec decPlain
+       validateDI infoPlain
+         DatatypeInfo
+           { datatypeName    = mkName "T43Plain"
+           , datatypeContext = []
+           , datatypeVars    = []
+           , datatypeVariant = Datatype
+           , datatypeCons    =
+               [ ConstructorInfo
+                   { constructorName       = mkName "MkT43Plain"
+                   , constructorVars       = []
+                   , constructorContext    = []
+                   , constructorFields     = []
+                   , constructorStrictness = []
+                   , constructorVariant    = NormalConstructor } ]
+           }
+
+       [decFam] <- [d| data instance T43Fam where  MkT43Fam :: T43Fam |]
+       infoFam  <- normalizeDec decFam
+       validateDI infoFam
+         DatatypeInfo
+           { datatypeName    = mkName "T43Fam"
+           , datatypeContext = []
+           , datatypeVars    = []
+           , datatypeVariant = DataInstance
+           , datatypeCons    =
+               [ ConstructorInfo
+                   { constructorName       = mkName "MkT43Fam"
+                   , constructorVars       = []
+                   , constructorContext    = []
+                   , constructorFields     = []
+                   , constructorStrictness = []
+                   , constructorVariant    = NormalConstructor } ]
+           }
+   )
+#endif
+
 #if MIN_VERSION_template_haskell(2,7,0)
 dataFamilyTest :: IO ()
 dataFamilyTest =
@@ -427,7 +475,7 @@ gadtFamTest =
                    , constructorVariant    = NormalConstructor }
                , ConstructorInfo
                    { constructorName       = '(:&&:)
-                   , constructorVars       = [PlainTV e]
+                   , constructorVars       = [KindedTV e starK]
                    , constructorContext    = [equalPred cTy (AppT ListT eTy)]
                    , constructorFields     = [eTy,dTy]
                    , constructorStrictness = [notStrictAnnot, notStrictAnnot]
@@ -453,7 +501,7 @@ gadtFamTest =
                    , constructorVariant    = NormalConstructor }
                , ConstructorInfo
                    { constructorName       = 'MkGadtFam5
-                   , constructorVars       = [PlainTV q]
+                   , constructorVars       = [KindedTV q starK]
                    , constructorContext    = [ equalPred cTy (ConT ''Bool)
                                              , equalPred dTy (ConT ''Bool)
                                              , equalPred qTy (ConT ''Char)
@@ -512,6 +560,16 @@ recordFamTest :: IO ()
 recordFamTest =
   $(do info <- reifyRecord 'famRec1
        validateCI info gadtRecFamCI)
+
+t46Test :: IO ()
+t46Test =
+  $(do info <- reifyDatatype 'MkT46
+       case info of
+         DatatypeInfo { datatypeCons = [ConstructorInfo { constructorContext = ctxt }]} ->
+           unless (null ctxt) (fail "regression test for ticket #46 failed")
+         _ -> fail "T46 should have exactly one constructor"
+       [| return () |])
+
 #endif
 
 fixityLookupTest :: IO ()
@@ -577,9 +635,7 @@ importedEqualityTest =
            , datatypeCons    =
                [ ConstructorInfo
                    { constructorName       = 'Refl
-                   , constructorVars       = [KindedTV k starK]
-                     -- This shouldn't happen, ideally. See #37.
-
+                   , constructorVars       = []
                    , constructorContext    = [equalPred a b]
                    , constructorFields     = []
                    , constructorStrictness = []
@@ -606,3 +662,101 @@ kindSubstTest =
        checkFreeVars substTy [k2]
        [| return () |])
 #endif
+
+#if __GLASGOW_HASKELL__ >= 800
+t37Test :: IO ()
+t37Test =
+  $(do infoA <- reifyDatatype ''T37a
+       let [k,a] = map (VarT . mkName) ["k","a"]
+       validateDI infoA
+         DatatypeInfo
+           { datatypeContext = []
+           , datatypeName    = ''T37a
+           , datatypeVars    = [SigT k starK, SigT a k]
+           , datatypeVariant = Datatype
+           , datatypeCons    =
+               [ ConstructorInfo
+                   { constructorName       = 'MkT37a
+                   , constructorVars       = []
+                   , constructorContext    = [equalPred k (ConT ''Bool)]
+                   , constructorFields     = []
+                   , constructorStrictness = []
+                   , constructorVariant    = NormalConstructor } ]
+           }
+
+       infoB <- reifyDatatype ''T37b
+       validateDI infoB
+         DatatypeInfo
+           { datatypeContext = []
+           , datatypeName    = ''T37b
+           , datatypeVars    = [SigT a k]
+           , datatypeVariant = Datatype
+           , datatypeCons    =
+               [ ConstructorInfo
+                   { constructorName       = 'MkT37b
+                   , constructorVars       = []
+                   , constructorContext    = [equalPred k (ConT ''Bool)]
+                   , constructorFields     = []
+                   , constructorStrictness = []
+                   , constructorVariant    = NormalConstructor } ]
+           }
+
+       infoC <- reifyDatatype ''T37c
+       validateDI infoC
+         DatatypeInfo
+           { datatypeContext = []
+           , datatypeName    = ''T37c
+           , datatypeVars    = [SigT a k]
+           , datatypeVariant = Datatype
+           , datatypeCons    =
+               [ ConstructorInfo
+                   { constructorName       = 'MkT37c
+                   , constructorVars       = []
+                   , constructorContext    = [equalPred a (ConT ''Bool)]
+                   , constructorFields     = []
+                   , constructorStrictness = []
+                   , constructorVariant    = NormalConstructor } ]
+           }
+   )
+
+polyKindedExTyvarTest :: IO ()
+polyKindedExTyvarTest =
+  $(do info <- reifyDatatype ''T48
+       let [a,x] = map mkName ["a","x"]
+       validateDI info
+         DatatypeInfo
+           { datatypeContext = []
+           , datatypeName    = ''T48
+           , datatypeVars    = [SigT (VarT a) starK]
+           , datatypeVariant = Datatype
+           , datatypeCons    =
+               [ ConstructorInfo
+                   { constructorName       = 'MkT48
+                   , constructorVars       = [KindedTV x (VarT a)]
+                   , constructorContext    = []
+                   , constructorFields     = [ConT ''Prox `AppT` VarT x]
+                   , constructorStrictness = [notStrictAnnot]
+                   , constructorVariant    = NormalConstructor } ]
+           }
+       -- Because validateCI uses a type variable substitution to normalize
+       -- away any alpha-renaming differences between constructors, it
+       -- unfortunately does not check if the uses of `a` in datatypeVars and
+       -- constructorVars are the same. We perform this check explicitly here.
+       case info of
+         DatatypeInfo { datatypeVars = [SigT (VarT a1) starK]
+                      , datatypeCons =
+                          [ ConstructorInfo
+                              { constructorVars = [KindedTV _ (VarT a2)] } ] } ->
+           unless (a1 == a2) $
+             fail $ "Two occurrences of the same variable have different names: "
+                 ++ show [a1, a2]
+       [| return () |]
+   )
+#endif
+
+regressionTest44 :: IO ()
+regressionTest44 =
+  $(do intToInt <- [t| Int -> Int |]
+       unified  <- unifyTypes [intToInt, intToInt]
+       unless (Map.null unified) (fail "regression test for ticket #44 failed")
+       [| return () |])
